/*
 * spruked.c
 *
 * Copyright (C) 2013  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2013  Karol Budniak <karol@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/sched.h>
#include <tmos/crc7k.h>

/*
 * We send data with SPRUKED_BUS_CYCLE period. Note that the MS_JIFFIES
 * converts miliseconds to jiffies with rouding up and this conversion
 * might increase period. For 1 ms period HZ=1000 is needed.
 */
#define SPRUKED_BUS_CYCLE	MS_JIFFIES(1)

#define SPRUKED_SYNC		1

static unsigned char addr;

#define PORT_LO		P0
#define PORT_HI		P2


/*
 * initial checksum - we start from 0x7f to avoid CRC problems with
 * inserting zeros before checksum - in this application it's not needed
 * but it's a common practice.
 */
#define CHECKSUM_INIT		0x7f

#ifndef CRC7K_POLY
#define CRC7K_POLY	0x76
#endif
static const unsigned int spd_table[] = {
	B9600, B19200, B38400, B115200
};

static void set_speed(void)
{
	unsigned char q = 0x00;

	P1 |= 0x18;
	q = (P1 & 0x18) >> 3;
	ser_set_spd(spd_table[q & 3]);
}

static unsigned char get_adress(void)
{
	unsigned char address = 0x00;

	P3 |= 0xE4;
	address = (P3 & 0xE0) >> 4;
	address = address | ((P3 & 0x04) >> 2);
	return ~address;
}

static void send_data(u16 x) __reentrant
{
	u8 hi = x >> 8;
	u8 lo = x;
	u8 csum = CHECKSUM_INIT;
	u8 t;

	t = (((hi & 1) << 1) | (lo & 1)) << 1;
	t |= addr | SPRUKED_SYNC;
	ser_txc(t);
	csum = crc7k_byte(csum, t);

	t = lo & 0xfe;
	ser_txc(t);
	csum = crc7k_byte(csum, t);

	t = hi & 0xfe;
	ser_txc(t);
	csum = crc7k_byte(csum, t);

	/*
	 * We use 7-bit CRC and we need to add single 0 bit here:
	 * multiple = (csum & 1) ? CRC_POLY : 0;
	 * csum = (csum >> 1) ^ multiple;
	 *
	 * Because we need to left shift CRC anyway we just do
	 * csum ^ (CRC_POLY << 1).
	 */
	if (csum & 1)
		csum ^= CRC7K_POLY << 1;

	ser_txc(~csum & 0xfe);
}

void app_main(void) __reentrant
{
	jiffies_t t;
	u16 x = 0;

	addr = get_adress() << 4;
	set_speed();

	t = get_jiffies();
	for (;;) {
		x = pack16(PORT_LO, PORT_HI);
		send_data(x);

		/*
		 * wait for next cycle
		 *
		 * this code assumes that we are faster than configured
		 * SPRUKED_BUS_CYCLE and we will wait here. Otherwise
		 * the time difference between current jiffies and
		 * expected wake-up jiffy t will increase and overflow.
		 */
		t += SPRUKED_BUS_CYCLE;
		while (!time_after_eq(get_jiffies(), t))
			schedule();
	}
}
